# -*- coding: utf-8 -*-
"""
Created on Wed Nov 21 11:12:32 2018

@author: Eidsvik
"""

import numpy as np
import matplotlib.pyplot as plt

"""
Voltage regulation - Stabilitet_del2_kap32
"""

saturation = False  #Imposes saturation effects
regulator_max = 3.5 #Determines max value for amplifier output
re_in = False       #Re-initialization strategy applied, true/false
strategy = True     #Enables step length adjustment strategy
halving_of_step_length = False  #Enables that the step length may not be reduced to more than half its previous value
fig_size = 100      #Determines dpi resolution for plots

if strategy == True:
    steps_before_h_change = 15
    k_step_length = 0.7
else:
    steps_before_h_change = 1
    k_step_length = 1.0
max_trunc_error = 1e-4

h_min = 0.00001
h_max = 0.05/k_step_length
h = [100*h_min, 100*h_min]
h_vector = [h[-1]]

t_curr = 0.0
t_vector = [t_curr]
t_final = 3.0
n = 4           #number of differential equations

"""Definition of constant(s)"""
#Regulator:
Kr = 3.0
Tr = 0.01
# regulator_max = 2.5

#Amplifier:
Ke = 3.0
Te = 0.15

#Generator:
K3 = 1.1
K6 = 1.1
Td0m = 0.3

#Measuring unit:
Tm = 0.02

"""Definition of reference set point(s)"""
#Generator output voltage:
du_ref = 1.0
du_ref_vector = [du_ref]

"""Differential equations"""
def f1(y): #Regulator output
    return ((1/Tr)*(Kr*(du_ref - y[3]) - y[0]))

def f2(y): #Amplifier output
    return ((1/Te)*(Ke*y[0] - y[1]))

def f3(y): #Generator output
    return ((1/(Td0m*K3))*(K3*K6*y[1] - y[2]))

def f4(y): #Measured generator output
    return ((1/Tm)*(y[2]-y[3]))

"""Predictor"""
def predictor(y, dy, d2y, h):
    p_y = y[-1] + h[1]*dy[-1] + ((h[1]**2)/2)*d2y[-1]
    p_dy = dy[-1] + h[1]*d2y[-1]
    p_d2y = d2y[-1]
    
    return p_y, p_dy, p_d2y

"""Corrector"""
def corrector(l1, l2, K2, h):
    
    L1 = np.array(n*[n*[0.0]])
    for i in range(0,n):
        for j in range(0,n):
            if i == j:
                L1[i][j] = l1
    
    dfdy = np.array([[-1/Tr, 0, 0, -Kr/Tr],[Ke/Te, -1/Te, 0, 0],[0.0, (K6*K3)/(Td0m*K3), -1/(Td0m*K3), 0],[0, 0, 1/Tm, -1/Tm]])
    A = L1 - h[-1]*dfdy
    
    """Initial estimate of Dy"""
    Dy_init1 = dy1[-1] * h[-1]
    Dy_init2 = dy2[-1] * h[-1]
    Dy_init3 = dy3[-1] * h[-1]
    Dy_init4 = dy4[-1] * h[-1]
    Dy = np.array([Dy_init1, Dy_init2, Dy_init3, Dy_init4])
    
    
    """Finding the Dy giving U = 0"""
    Dy = newton_raphson(A, Dy, l1)
    
    """Updating y1, y2 and their derivatives"""
    y1_next = p_y1[-1] + Dy[0]                                             #calculates next value of y
    dy1_next = (h[-1]*p_dy1[-1] + l1*Dy[0]) / h[-1]                         #calculates next value of first derivative of y
    d2y1_next = (((h[-1]**2)/2)*p_d2y1[-1] + l2*Dy[0]) / ((h[-1]**2)/2)    #calculates next value of second derivative of y
    
    y2_next = p_y2[-1] + Dy[1]
    dy2_next = (h[-1]*p_dy2[-1] + l1*Dy[1]) / h[-1]
    d2y2_next = (((h[-1]**2)/2)*p_d2y2[-1] + l2*Dy[1]) / ((h[-1]**2)/2)
    
    y3_next = p_y3[-1] + Dy[2]
    dy3_next = (h[-1]*p_dy3[-1] + l1*Dy[2]) / h[-1]
    d2y3_next = (((h[-1]**2)/2)*p_d2y3[-1] + l2*Dy[2]) / ((h[-1]**2)/2)
    
    y4_next = p_y4[-1] + Dy[3]
    dy4_next = (h[-1]*p_dy4[-1] + l1*Dy[3]) / h[-1]
    d2y4_next = (((h[-1]**2)/2)*p_d2y4[-1] + l2*Dy[3]) / ((h[-1]**2)/2)

    
    return y1_next, dy1_next, d2y1_next, y2_next, dy2_next, d2y2_next, y3_next, dy3_next, d2y3_next, y4_next, dy4_next, d2y4_next, Dy, dfdy

"""NR method, auxiliary calculations for corrector"""
def newton_raphson(A, Dy, l1):
    error = 1e-5
#    for k in range(2):
    """Iterating to find Dy that gives U(Dy) = 0"""
    U_mismatch = np.array([error + 1.0])
    while max(abs(U_mismatch)) > error:
        
        """Calculation of U"""
        U1 = h[-1]*p_dy1[-1] + l1*Dy[0] - h[-1]*f1([(Dy[0] + p_y1[-1]), (Dy[1] + p_y2[-1]), (Dy[2] + p_y3[-1]), (Dy[3] + p_y4[-1])])
        U2 = h[-1]*p_dy2[-1] + l1*Dy[1] - h[-1]*f2([(Dy[0] + p_y1[-1]), (Dy[1] + p_y2[-1]), (Dy[2] + p_y3[-1]), (Dy[3] + p_y4[-1])])
        U3 = h[-1]*p_dy3[-1] + l1*Dy[2] - h[-1]*f3([(Dy[0] + p_y1[-1]), (Dy[1] + p_y2[-1]), (Dy[2] + p_y3[-1]), (Dy[3] + p_y4[-1])])
        U4 = h[-1]*p_dy4[-1] + l1*Dy[3] - h[-1]*f4([(Dy[0] + p_y1[-1]), (Dy[1] + p_y2[-1]), (Dy[2] + p_y3[-1]), (Dy[3] + p_y4[-1])])
        U_mismatch = np.array([-U1, -U2, -U3, -U4])
        
        D_Dy = np.linalg.solve(A, U_mismatch)          #Calculates how much we need to add to Dy to get 0 on RHS
        
        Dy += D_Dy

    return Dy
        
#        print("After NR: U_mismatch =\n", U_mismatch)

"""Definition of constants l1, l2, K2"""
def constants(h):
    """Calculates constants l1, l2 and K2 based on step length h"""
    l1 = (2*h[1] + h[0])/(h[1] + h[0])
    l2 = h[1] / (h[1] + h[0])
    K2 = (1/6) * (((h[1] + h[0])**2) / (h[1] * (2*h[1] + h[0])))
    
    return l1, l2, K2


"""
Beginning of algorithm
"""

"""Initial values"""
#Regulator output:
y1_0 = 2.7

#Amplifier output:
y2_0 = -2.2

#Generator output:
y3_0 = 0.0

#Measured generator output:
y4_0 = y3_0

dy1_0 = f1([y1_0, y2_0, y3_0, y4_0])      #initial value of first derivative of y1
dy2_0 = f2([y1_0, y2_0, y3_0, y4_0])      #initial value of first derivative of y2
dy3_0 = f3([y1_0, y2_0, y3_0, y4_0])      #initial value of first derivative of y3
dy4_0 = f4([y1_0, y2_0, y3_0, y4_0])

d2y1_0 = 0.0                        #initial value of second derivative of y1
d2y2_0 = 0.0                        #initial value of second derivative of y2
d2y3_0 = 0.0                        #initial value of second derivative of y3
d2y4_0 = 0.0

y1 = [y1_0]
dy1 = [dy1_0]
d2y1 = [d2y1_0]

y2 = [y2_0]
dy2 = [dy2_0]
d2y2 = [d2y2_0]

y3 = [y3_0]
dy3 = [dy3_0]
d2y3 = [d2y3_0]

y4 = [y4_0]
dy4 = [dy4_0]
d2y4 = [d2y4_0]

p_y1 = []
p_dy1 = []
p_d2y1 = []

p_y2 = []
p_dy2 = []
p_d2y2 = []

p_y3 = []
p_dy3 = []
p_d2y3 = []

p_y4 = []
p_dy4 = []
p_d2y4 = []

increase_counter = 0
decrease_counter = 0
counter = 0
sat_counter = 0 #to be used to determine when saturation ends

while t_curr < t_final:
    du_ref_vector.append(du_ref)

    """Predictor phase"""
    p_y1_f, p_dy1_f, p_d2y1_f = predictor(y1, dy1, d2y1, h)
    p_y2_f, p_dy2_f, p_d2y2_f = predictor(y2, dy2, d2y2, h)
    p_y3_f, p_dy3_f, p_d2y3_f = predictor(y3, dy3, d2y3, h)
    p_y4_f, p_dy4_f, p_d2y4_f = predictor(y4, dy4, d2y4, h)
    
    p_y1.append(p_y1_f)
    p_dy1.append(p_dy1_f)
    p_d2y1.append(p_d2y1_f)
    
    p_y2.append(p_y2_f)
    p_dy2.append(p_dy2_f)
    p_d2y2.append(p_d2y2_f)
    
    p_y3.append(p_y3_f)
    p_dy3.append(p_dy3_f)
    p_d2y3.append(p_d2y3_f)
    
    p_y4.append(p_y4_f)
    p_dy4.append(p_dy4_f)
    p_d2y4.append(p_d2y4_f)
    
    
    """Corrector phase"""
    l1, l2, K2 = constants(h)
    y1_c, dy1_c, d2y1_c, y2_c, dy2_c, d2y2_c, y3_c, dy3_c, d2y3_c, y4_c, dy4_c, d2y4_c, Dy, dfdy = corrector(l1, l2, K2, h)
    #y1_c = candidate for y1, dy1_c = candidate for dy1 etc.
    #d = first order derivative
    #d2 = second order derivative
    t_curr += h[-1]
    t_vector.append(t_curr)
    
    counter += 1
    h_vector.append(h[-1])
    
    max_trunc_error_var = 0.0
    trunc_error = K2 * 2 * l2 * max(abs(Dy))
    correction_counter = 0
    h_has_been_updated = False
    if counter % steps_before_h_change == 0 and t_curr > 0.0:
        if trunc_error > max_trunc_error or trunc_error < 0.5*max_trunc_error:
            h_temp = h[1]
            if strategy == True:
                h[1] = k_step_length*min(h_max, 2*h_temp, ((max_trunc_error/trunc_error)**0.5)*h_temp)
            elif strategy == False:
                h[1] = ((max_trunc_error/trunc_error)**0.5)*h_temp
            h[0] = h_temp
            y1_c, dy1_c, d2y1_c, y2_c, dy2_c, d2y2_c, y3_c, dy3_c, d2y3_c, y4_c, dy4_c, d2y4_c, Dy, dfdy = corrector(l1, l2, K2, h)
            trunc_error = K2 * 2 * l2 * max(abs(Dy))
            correction_counter += 1
            h_has_been_updated = True
            
            if h[1] < h[0]:
                decrease_counter += 1   #logs number of times h has been decreased
            else:
                increase_counter += 1   #logs number of times h has been increased
    if h[1] > h_max:
        h[1] = h_max
    elif h[1] < h_min:
        h[1] = h_min
    if halving_of_step_length:
        if h[1] < 0.5*h[0]:
            h[1] = 0.5*h[0]

    """Saturation"""
    if saturation == True:
        if y2_c > regulator_max:
            y2_c = regulator_max
            dy2_c = 0.0
            d2y2_c = 0.0
            sat_counter = counter
        elif y2_c < regulator_max and sat_counter > 0 and (counter - sat_counter) == 1 and re_in == True:
            print("RE-INIT")
            dy1_c = f1([y1[-1], y2[-1], y3[-1], y4[-1]])
            dy2_c = f2([y1[-1], y2[-1], y3[-1], y4[-1]])
            dy3_c = f3([y1[-1], y2[-1], y3[-1], y4[-1]])
            dy4_c = f4([y1[-1], y2[-1], y3[-1], y4[-1]])
            
            d2y1_c, d2y2_c, d2y3_c, d2y4_c = dfdy.dot(np.array([dy1_c, dy2_c, dy3_c, dy4_c]))

    y1.append(y1_c)
    dy1.append(dy1_c)
    d2y1.append(d2y1_c)
    
    y2.append(y2_c)
    dy2.append(dy2_c)
    d2y2.append(d2y2_c)

    y3.append(y3_c)
    dy3.append(dy3_c)
    d2y3.append(d2y3_c)
    
    y4.append(y4_c)
    dy4.append(dy4_c)
    d2y4.append(d2y4_c)
    
        
def make_pred_vector_long_enough(): #The vector of predicted values is 1 element shorter than the vector with actual values
    p_y1.append(p_y1_f)
    p_dy1.append(p_dy1_f)
    p_d2y1.append(p_d2y1_f)
    
    p_y2.append(p_y2_f)
    p_dy2.append(p_dy2_f)
    p_d2y2.append(p_d2y2_f)
    
    p_y3.append(p_y3_f)
    p_dy3.append(p_dy3_f)
    p_d2y3.append(p_d2y3_f)
    
    p_y4.append(p_y4_f)
    p_dy4.append(p_dy4_f)
    p_d2y4.append(p_d2y4_f)
make_pred_vector_long_enough()

"""Aspects of h"""
h_average = sum(h_vector)/(float(len(h_vector)))
h_maximum = max(h_vector)
h_minimum = min(h_vector)

def plotting():
    """Plotting"""
    plt.figure(dpi = fig_size)
    plt.xlabel("Time [sec]")
    plt.ylabel("State variables")
    plt.grid()
    plt.plot(t_vector, y1, 'b', label="Regulator output Dy")
    plt.plot(t_vector, y2, 'r', label="Amplifier output Du_df")
    plt.plot(t_vector, y3, 'r--', label="Generator output D_u")
    plt.plot(t_vector, y4, 'b--', label = "Measured gen. output Du_m")
    plt.plot(t_vector, du_ref_vector, "k--", label="Reference")
    plt.legend(loc="best")
    
    # plt.figure(dpi = fig_size)
    # plt.xlabel("Time [sec]")
    # plt.ylabel("Amplifier ouput first derivative")
    # plt.grid()
    # plt.plot(t_vector, dy2, 'b', label="Amplifier output 1st der")
    # plt.legend(loc="best")
    
    # plt.figure(dpi = fig_size)
    # plt.xlabel("Time [sec]")
    # plt.ylabel("Amplifier ouput first derivative")
    # plt.grid()
    # plt.plot(t_vector, d2y2, 'b', label="Amplifier output 2der")
    # plt.legend(loc="best")
    
    # plt.figure(dpi = fig_size)
    # plt.xlabel("Time [sec]")
    # plt.ylabel("Step length [sec]")
    # plt.grid()
    # plt.plot(t_vector, h_vector, "b", label="Step length")
    # #plt.plot(t_vector, h_ma, "r", label="500 element moving average")
    # #plt.plot(t_vector, h_ma2, "r--", label="1000 element moving average")
    # plt.legend(loc="best")
def end_of_algorithm_stats():
    if saturation:
        print("Saturated case.")
    else:
        print("Non-saturated case.")
    print("Algorithm finished in %d iterations." % counter)
    print("The step length is allowed to change every %d steps." % steps_before_h_change)
    print("Increased step length %d times." % increase_counter)
    print("Decreased step length %d times." % decrease_counter)
    print("Step length adjustments: %d." % (decrease_counter + increase_counter))
    print("The average step length is %0.6fs." % h_average)
    print("The maximum step length is %0.6fs." % h_maximum)
    print("The minimum step length is %0.6fs." % h_minimum)
    print("Permitted step length max: %0.7fs, min: %0.7fs" % (h_max*k_step_length, h_min))

plotting()
end_of_algorithm_stats()